# -*- coding: utf-8 -*-

"""Non-graphical part of the Energy step in a TorchANI flowchart"""

import logging
from pathlib import Path
import pkg_resources
import pprint  # noqa: F401
import textwrap

from tabulate import tabulate

import torchani_step
import molsystem
import seamm
from seamm_util import ureg, Q_  # noqa: F401
import seamm_util.printing as printing
from seamm_util.printing import FormattedText as __

# In addition to the normal logger, two logger-like printing facilities are
# defined: "job" and "printer". "job" send output to the main job.out file for
# the job, and should be used very sparingly, typically to echo what this step
# will do in the initial summary of the job.
#
# "printer" sends output to the file "step.out" in this steps working
# directory, and is used for all normal output from this step.

logger = logging.getLogger(__name__)
job = printing.getPrinter()
printer = printing.getPrinter("TorchANI")

# Add this module's properties to the standard properties
path = Path(pkg_resources.resource_filename(__name__, "data/"))
csv_file = path / "properties.csv"
if path.exists():
    molsystem.add_properties_from_file(csv_file)


class Energy(seamm.Node):
    """
    The non-graphical part of a Energy step in a flowchart.

    Attributes
    ----------
    parser : configargparse.ArgParser
        The parser object.

    options : tuple
        It contains a two item tuple containing the populated namespace and the
        list of remaining argument strings.

    subflowchart : seamm.Flowchart
        A SEAMM Flowchart object that represents a subflowchart, if needed.

    parameters : EnergyParameters
        The control parameters for Energy.

    See Also
    --------
    TkEnergy,
    Energy, EnergyParameters
    """

    def __init__(self, flowchart=None, title="Energy", extension=None, logger=logger):
        """A substep for Energy in a subflowchart for TorchANI.

        You may wish to change the title above, which is the string displayed
        in the box representing the step in the flowchart.

        Parameters
        ----------
        flowchart: seamm.Flowchart
            The non-graphical flowchart that contains this step.

        title: str
            The name displayed in the flowchart.
        extension: None
            Not yet implemented
        logger : Logger = logger
            The logger to use and pass to parent classes

        Returns
        -------
        None
        """
        logger.debug(f"Creating Energy {self}")

        super().__init__(
            flowchart=flowchart,
            title=title,
            extension=extension,
            module=__name__,
            logger=logger,
        )  # yapf: disable

        self._calculation = "energy"
        self._model = None
        self._metadata = torchani_step.metadata
        self.parameters = torchani_step.EnergyParameters()

    @property
    def header(self):
        """A printable header for this section of output"""
        return "Step {}: {}".format(".".join(str(e) for e in self._id), self.title)

    @property
    def version(self):
        """The semantic version of this module."""
        return torchani_step.__version__

    @property
    def git_revision(self):
        """The git version of this module."""
        return torchani_step.__git_revision__

    def description_text(self, P=None):
        """Create the text description of what this step will do.
        The dictionary of control values is passed in as P so that
        the code can test values, etc.

        Parameters
        ----------
        P: dict
            An optional dictionary of the current values of the control
            parameters.
        Returns
        -------
        str
            A description of the current step.
        """
        if not P:
            P = self.parameters.values_to_dict()

        if P["gradients"]:
            text = "Calculating the energy and gradients "
        else:
            text = "Calculating the energy "
        text += "using the ANI machine learning model {model}."
        submodels = P["submodel"]
        if submodels == "all":
            text += (
                " All the parameterizations of the model will be used, and the "
                "results averaged."
            )
        else:
            if "," in submodels or "-" in submodels or self.is_expr(submodels):
                text += " These parameterizations of the model will be used: "
                "{submodels}, and the results will be averaged."
            else:
                text += " The {submodels} parameterization of the model will be used."

        return self.header + "\n" + __(text, **P, indent=4 * " ").__str__()

    def get_input(self, schema):
        """Get the input for the energy in TorchANI.

        Parameters
        ----------
        None

        Returns
        -------
        seamm.Node
            The next node object in the flowchart.
        """
        # Create the directory
        directory = Path(self.directory)
        directory.mkdir(parents=True, exist_ok=True)

        # Get the values of the parameters, dereferencing any variables
        P = self.parameters.current_values_to_dict(
            context=seamm.flowchart_variables._data
        )

        # Get the current system and configuration (ignoring the system...)
        _, configuration = self.get_system_configuration(None)

        # Set up the description.
        self.description = []
        self.description.append(__(self.description_text(P), **P, indent=self.indent))

        # Check the elements
        parameterization = P["model"]
        if parameterization == "ANI-1x":
            covered_elements = {"C", "H", "N", "O"}
        elif parameterization == "ANI-1ccx":
            covered_elements = {"C", "H", "N", "O"}
        elif parameterization == "ANI-2x":
            covered_elements = {"C", "H", "N", "O", "F", "S", "Cl"}
        else:
            raise RuntimeError(f"Don't recognize ANI model '{parameterization}'.")

        elements = set(configuration.atoms.symbols)
        if not elements <= covered_elements:
            raise RuntimeError(
                f"The {parameterization} parameterization covers the "
                f" elements {sorted(covered_elements)}. The system has "
                f"{sorted(elements)}."
            )

        # Get the QCSchema structure for the molecule
        schema = configuration.to_cms_schema()

        # add workflow step to the schema
        step = {
            "model": {
                "method": "ML",
                "model": "ANI",
                "parameterization": P["model"],
            },
            "keywords": {"submodel": P["submodel"]},
            "provenance": {
                "creator": "SEAMM/torchani_step",
                "version": "1.1",
                "routine": "torchani_step.energy.get_input",
            },
            "required results": [
                "energy",
            ],
        }

        results = step["required results"]
        if P["gradients"]:
            results.append("gradients")

        schema["workflow"] = [step]

        # Add other citations here or in the appropriate place in the code.
        # Add the bibtex to data/references.bib, and add a self.reference.cite
        # similar to the above to actually add the citation to the references.

        return schema

    def analyze(self, indent="", **kwargs):
        """Do any analysis of the output from this step.

        Also print important results to the local step.out file using
        "printer".

        Parameters
        ----------
        indent: str
            An extra indentation for the output
        """
        P = self.parameters.current_values_to_dict(
            context=seamm.flowchart_variables._data
        )

        # Get the appropriate system/configuration for the new coordinates
        system, configuration = self.get_system_configuration(None)

        schema = kwargs["schema"]
        step_no = kwargs["step_no"]

        table = {
            "System": [],
            "Configuration": [],
            "Energy": [],
            "Std Dev": [],
        }

        # The model chemistry
        self.model = f"ANI/{P['model']}"

        have_stdev = False
        for system in schema["systems"]:
            system_name = system["name"]
            for no, config_data in enumerate(system["configurations"]):
                if no == 0:
                    table["System"].append(system_name)
                else:
                    table["System"].append("")
                table["Configuration"].append(config_data["name"])
                results = config_data["results"]["data"][step_no]
                results["model"] = "TorchANI/" + self.model
                results["energy,units"] = "E_h"
                results["gradients,units"] = "E_h/Å"

                table["Energy"].append(f"{results['energy']:.6f}")
                if "energy, stdev" in results:
                    table["Std Dev"].append(f"{results['energy, stdev']:.6f}")
                    have_stdev = True
                else:
                    table["Std Dev"].append("")

                # And store results
                self.store_results(
                    configuration=configuration,
                    data=results,
                )

        if not have_stdev:
            del table["Std Dev"]

        text_lines = []
        text_lines.append("                     Results")
        text_lines.append(
            tabulate(
                table,
                headers="keys",
                tablefmt="rounded_outline",
                colalign=("left", "left", "decimal", "decimal"),
                disable_numparse=True,
            )
        )
        text_lines.append("\n")

        text = textwrap.indent("\n".join(text_lines), self.indent + 4 * " ")
        printer.normal(text)

        # Citation for the specific method
        self.references.cite(
            raw=self._bibliography[P["model"]],
            alias="P['model']",
            module="torchani_step",
            level=2,
            note=f"The citation for the {P['model']} ML model.",
        )



# -*- coding: utf-8 -*-

"""Non-graphical part of the TorchANI step in a SEAMM flowchart"""

import configparser
import importlib
import json
import logging
from pathlib import Path
import pkg_resources
import pprint  # noqa: F401
import shutil
import sys

import torchani_step
import molsystem
import seamm
import seamm_util
from seamm_util import ureg, Q_, CompactJSONEncoder  # noqa: F401
import seamm_util.printing as printing
from seamm_util.printing import FormattedText as __

# In addition to the normal logger, two logger-like printing facilities are
# defined: "job" and "printer". "job" send output to the main job.out file for
# the job, and should be used very sparingly, typically to echo what this step
# will do in the initial summary of the job.
#
# "printer" sends output to the file "step.out" in this steps working
# directory, and is used for all normal output from this step.

logger = logging.getLogger(__name__)
job = printing.getPrinter()
printer = printing.getPrinter("TorchANI")

# Add this module's properties to the standard properties
path = Path(pkg_resources.resource_filename(__name__, "data/"))
csv_file = path / "properties.csv"
if path.exists():
    molsystem.add_properties_from_file(csv_file)


class TorchANI(seamm.Node):
    """
    The non-graphical part of a TorchANI step in a flowchart.

    Attributes
    ----------
    parser : configargparse.ArgParser
        The parser object.

    options : tuple
        It contains a two item tuple containing the populated namespace and the
        list of remaining argument strings.

    subflowchart : seamm.Flowchart
        A SEAMM Flowchart object that represents a subflowchart, if needed.

    parameters : TorchANIParameters
        The control parameters for TorchANI.

    See Also
    --------
    TkTorchANI,
    TorchANI, TorchANIParameters
    """

    def __init__(
        self,
        flowchart=None,
        title="TorchANI",
        namespace="org.molssi.seamm.torchani",
        extension=None,
        logger=logger,
    ):
        """A step for TorchANI in a SEAMM flowchart.

        You may wish to change the title above, which is the string displayed
        in the box representing the step in the flowchart.

        Parameters
        ----------
        flowchart: seamm.Flowchart
            The non-graphical flowchart that contains this step.

        title: str
            The name displayed in the flowchart.
        namespace : str
            The namespace for the plug-ins of the subflowchart
        extension: None
            Not yet implemented
        logger : Logger = logger
            The logger to use and pass to parent classes

        Returns
        -------
        None
        """
        logger.debug(f"Creating TorchANI {self}")
        self.subflowchart = seamm.Flowchart(
            parent=self, name="TorchANI", namespace=namespace
        )  # yapf: disable

        super().__init__(
            flowchart=flowchart,
            title="TorchANI",
            extension=extension,
            module=__name__,
            logger=logger,
        )  # yapf: disable

        self._metadata = torchani_step.metadata
        self.parameters = torchani_step.TorchANIParameters()

    @property
    def version(self):
        """The semantic version of this module."""
        return torchani_step.__version__

    @property
    def git_revision(self):
        """The git version of this module."""
        return torchani_step.__git_revision__

    def set_id(self, node_id):
        """Set the id for node to a given tuple"""
        self._id = node_id

        # and set our subnodes
        self.subflowchart.set_ids(self._id)

        return self.next()

    def create_parser(self):
        """Setup the command-line / config file parser"""
        parser_name = self.step_type
        parser = seamm_util.getParser(name="SEAMM")

        # Remember if the parser exists ... this type of step may have been
        # found before
        parser_exists = parser.exists(parser_name)

        # Create the standard options, e.g. log-level
        result = super().create_parser(name=parser_name)

        if parser_exists:
            return result

        return result

    def description_text(self, P=None):
        """Create the text description of what this step will do.
        The dictionary of control values is passed in as P so that
        the code can test values, etc.

        Parameters
        ----------
        P: dict
            An optional dictionary of the current values of the control
            parameters.
        Returns
        -------
        str
            A description of the current step.
        """
        self.subflowchart.root_directory = self.flowchart.root_directory

        # Get the first real node
        node = self.subflowchart.get_node("1").next()

        text = self.header + "\n\n"
        while node is not None:
            try:
                text += __(node.description_text(), indent=3 * " ").__str__()
            except Exception as e:
                print(f"Error describing torchani flowchart: {e} in {node}")
                self.logger.critical(
                    f"Error describing torchani flowchart: {e} in {node}"
                )
                raise
            except:  # noqa: E722
                print(
                    "Unexpected error describing torchani flowchart: {} in {}".format(
                        sys.exc_info()[0], str(node)
                    )
                )
                self.logger.critical(
                    "Unexpected error describing torchani flowchart: {} in {}".format(
                        sys.exc_info()[0], str(node)
                    )
                )
                raise
            text += "\n"
            node = node.next()

        return text

    def run(self):
        """Run a TorchANI step.

        Parameters
        ----------
        None

        Returns
        -------
        seamm.Node
            The next node object in the flowchart.
        """
        # Create the directory
        directory = Path(self.directory)
        directory.mkdir(parents=True, exist_ok=True)

        next_node = super().run(printer)

        # Print our header to the main output
        printer.important(self.header)
        printer.important("")

        # Access the options and find the executable
        seamm_options = self.global_options

        # Get the first real node
        node1 = self.subflowchart.get_node("1").next()

        # Print what we will do as we get the input
        schema = {}
        node = node1
        nodes = []
        while node is not None:
            nodes.append(node)
            schema = node.get_input(schema)
            for value in node.description:
                printer.important(value)
                printer.important(" ")
            node = node.next()

        schema_name = schema["schema name"]
        schema_version = schema["schema version"]
        input_data = f"!MolSSI {schema_name} {schema_version}\n"
        input_data += json.dumps(
            schema, indent=4, cls=CompactJSONEncoder, sort_keys=True
        )
        files = {"input.json": input_data}
        self.logger.debug("input.json:\n" + files["input.json"])
        executor = self.flowchart.executor

        # Read configuration file for TorchANI if it exists
        executor_type = executor.name
        full_config = configparser.ConfigParser()
        ini_dir = Path(seamm_options["root"]).expanduser()
        path = ini_dir / "torchani.ini"

        if path.exists():
            full_config.read(ini_dir / "torchani.ini")

        # If the section we need doesn't exist, get the default
        if not path.exists() or executor_type not in full_config:
            resources = importlib.resources.files("torchani_step") / "data"
            ini_text = (resources / "torchani.ini").read_text()
            full_config.read_string(ini_text)

        # Getting desperate! Look for an executable in the path
        if executor_type not in full_config:
            path = shutil.which("SEAMM_TorchANI.py")
            if path is None:
                raise RuntimeError(
                    f"No section for '{executor_type}' in TorchANI ini file "
                    f"({ini_dir / 'torchani.ini'}), nor in the defaults, nor "
                    "in the path!"
                )
            else:
                full_config[executor_type] = {
                    "installation": "local",
                    "code": str(path),
                }

        # If the ini file does not exist, write it out!
        if not path.exists():
            with path.open("w") as fd:
                full_config.write(fd)
            printer.normal(f"Wrote the TorchANI configuration file to {path}")
            printer.normal("")

        config = dict(full_config.items(executor_type))
        # Use the matching version of the seamm-torchani image by default.
        config["version"] = self.version

        cmd = ["{code}", "input.json", ">", "output.txt", "2>", "stderr.txt"]

        return_files = [
            "output.json",
            "output.txt",
            "stderr.txt",
        ]

        self.logger.debug(f"{cmd=}")

        result = executor.run(
            cmd=cmd,
            config=config,
            directory=self.directory,
            files=files,
            return_files=return_files,
            in_situ=True,
            shell=True,
        )

        if not result:
            self.logger.error("There was an error running TorchANI")
            return None

        self.logger.debug("\n" + pprint.pformat(result))

        self.logger.debug("stdout:\n" + result["stdout"])
        if "stderr.txt" in result and "data" in result["stderr.txt"]:
            if result["stderr.txt"]["data"] != "":
                lines = result["stderr.txt"]["data"].splitlines()
                tmp = []
                for line in lines:
                    if "cuaev not installed" in line:
                        continue
                    if "Creating a tensor from a list of numpy" in line:
                        continue
                    if "cell = torch.tensor(self.atoms.get_cell(complete=True)" in line:
                        continue
                    tmp.append(line)
                if len(tmp) > 0:
                    self.logger.warning("stderr:\n" + "\n".join(tmp))

        lines = result["output.json"]["data"].splitlines()
        line = lines[0]
        if line[0:7] != "!MolSSI":
            raise RuntimeError(
                "Output file is not a MolSSI schema file, organization is not MolSSI: "
                f"'{line}'"
            )
        tmp = line.split()
        if len(tmp) < 3:
            raise RuntimeError(f"Output file is not a MolSSI schema file: '{line}'")
        if tmp[1] != "cms_schema":
            raise RuntimeError(f"Output file is not a CMS schema file: '{line}'")

        schema = json.loads("\n".join(lines[1:]))

        # Check that the job ran OK
        for step_no, step in enumerate(schema["workflow"]):
            if "success" not in step:
                self.logger.warning(f"Step {step_no} did not run")
                continue
            if step["success"]:
                nodes[step_no].analyze(schema=schema, step_no=step_no)
            else:
                if "error" in step:
                    self.logger.error("TorchANI had an error:\n\n" + step["error"])
                    raise RuntimeError("TorchANI had an error:\n\n" + step["error"])

        # Add other citations here or in the appropriate place in the code.
        # Add the bibtex to data/references.bib, and add a self.reference.cite
        # similar to the above to actually add the citation to the references.
        self.references.cite(
            raw=self._bibliography["TorchANI"],
            alias="TorchANI",
            module="torchani_step",
            level=1,
            note="The citation for the TorchANI software.",
        )
        self.references.cite(
            raw=self._bibliography["ANI"],
            alias="ANI",
            module="torchani_step",
            level=1,
            note="The citation for the ANI ML.",
        )
        self.references.cite(
            raw=self._bibliography["ANI_dataset"],
            alias="ANI_dataset",
            module="torchani_step",
            level=2,
            note="The citation for the ANI dataset.",
        )

        return next_node



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Tests for `torchani_step` package."""

import pytest  # noqa: F401
import torchani_step  # noqa: F401


def test_construction():
    """Just create an object and test its type."""
    result = torchani_step.TorchANI()
    assert str(type(result)) == "<class 'torchani_step.torchani.TorchANI'>"



# -*- coding: utf-8 -*-

"""Non-graphical part of the Optimization step in a TorchANI flowchart"""

import logging
import math
from pathlib import Path
import pkg_resources
import pprint  # noqa: F401
import textwrap

from tabulate import tabulate

import torchani_step
import molsystem
import seamm
from seamm_util import ureg, Q_  # noqa: F401
import seamm_util.printing as printing
from seamm_util.printing import FormattedText as __

# In addition to the normal logger, two logger-like printing facilities are
# defined: "job" and "printer". "job" send output to the main job.out file for
# the job, and should be used very sparingly, typically to echo what this step
# will do in the initial summary of the job.
#
# "printer" sends output to the file "step.out" in this steps working
# directory, and is used for all normal output from this step.

logger = logging.getLogger(__name__)
job = printing.getPrinter()
printer = printing.getPrinter("TorchANI")

# Add this module's properties to the standard properties
path = Path(pkg_resources.resource_filename(__name__, "data/"))
csv_file = path / "properties.csv"
if path.exists():
    molsystem.add_properties_from_file(csv_file)


class Optimization(torchani_step.Energy):
    """
    The non-graphical part of a Optimization step in a flowchart.

    Attributes
    ----------
    parser : configargparse.ArgParser
        The parser object.

    options : tuple
        It contains a two item tuple containing the populated namespace and the
        list of remaining argument strings.

    subflowchart : seamm.Flowchart
        A SEAMM Flowchart object that represents a subflowchart, if needed.

    parameters : OptimizationParameters
        The control parameters for Optimization.

    See Also
    --------
    TkOptimization,
    Optimization, OptimizationParameters
    """

    def __init__(
        self, flowchart=None, title="Optimization", extension=None, logger=logger
    ):
        """A substep for Optimization in a subflowchart for TorchANI.

        You may wish to change the title above, which is the string displayed
        in the box representing the step in the flowchart.

        Parameters
        ----------
        flowchart: seamm.Flowchart
            The non-graphical flowchart that contains this step.

        title: str
            The name displayed in the flowchart.
        extension: None
            Not yet implemented
        logger : Logger = logger
            The logger to use and pass to parent classes

        Returns
        -------
        None
        """
        logger.debug(f"Creating Optimization {self}")

        super().__init__(
            flowchart=flowchart,
            title=title,
            extension=extension,
            logger=logger,
        )

        self._calculation = "optimization"
        self._model = None
        self._metadata = torchani_step.metadata
        self.parameters = torchani_step.OptimizationParameters()

    @property
    def header(self):
        """A printable header for this section of output"""
        return "Step {}: {}".format(".".join(str(e) for e in self._id), self.title)

    @property
    def version(self):
        """The semantic version of this module."""
        return torchani_step.__version__

    @property
    def git_revision(self):
        """The git version of this module."""
        return torchani_step.__git_revision__

    def description_text(self, P=None):
        """Create the text description of what this step will do.
        The dictionary of control values is passed in as P so that
        the code can test values, etc.

        Parameters
        ----------
        P: dict
            An optional dictionary of the current values of the control
            parameters.
        Returns
        -------
        str
            A description of the current step.
        """
        if not P:
            P = self.parameters.values_to_dict()

        text = "Optimizing the structure using the ANI machine learning model {model}."
        submodels = P["submodel"]
        if submodels == "all":
            text += (
                " All the parameterizations of the model will be used, and the "
                "results averaged."
            )
        else:
            if "," in submodels or "-" in submodels or self.is_expr(submodels):
                text += " These parameterizations of the model will be used: "
                "{submodels}, and the results will be averaged."
            else:
                text += " The {submodels} parameterization of the model will be used."

        text += "\n\nThe optimization will use the {minimizer}"
        if "minimizer" not in P["minimizer"]:
            text += " minimizer"
        text += (
            " with a convergence criterion of {convergence} and a limit of {max steps}"
            " steps."
        )

        text += "\n"
        text += seamm.standard_parameters.structure_handling_description(P)

        return self.header + "\n" + __(text, **P, indent=4 * " ").__str__()

    def get_input(self, schema):
        """Get the input for the optimization in TorchANI.

        Parameters
        ----------
        None

        Returns
        -------
        seamm.Node
            The next node object in the flowchart.
        """
        # Create the directory
        directory = Path(self.directory)
        directory.mkdir(parents=True, exist_ok=True)

        # Get the values of the parameters, dereferencing any variables
        P = self.parameters.current_values_to_dict(
            context=seamm.flowchart_variables._data
        )

        # Get the schema for the energy of the structure
        schema = super().get_input(schema)

        results = schema["workflow"][0]["required results"]
        if "gradients" not in results:
            results.append("gradients")
        results.append("optimized structure")

        schema["control parameters"] = {
            "optimization": {
                "minimizer": P["minimizer"],
                "maximum steps": P["max steps"],
                "convergence": P["convergence"].magnitude,
                "convergence units": str(P["convergence"].units),
            }
        }

        # Set up the description, overwriting that of the energy.
        self.description = []
        self.description.append(__(self.description_text(P), **P, indent=self.indent))

        return schema

    def analyze(self, indent="", schema=None, table=None, step_no=None, **kwargs):
        """Do any analysis of the output from this step.

        Also print important results to the local step.out file using
        "printer".

        Parameters
        ----------
        indent: str
            An extra indentation for the output
        """
        if schema is None:
            printer.normal(
                __(
                    "The Optimization step failed. There is no output at all!",
                    indent=4 * " ",
                    wrap=True,
                    dedent=False,
                )
            )
            return

        if "workflow" not in schema or len(schema["workflow"]) == 0:
            printer.normal(
                __(
                    "The Optimization step failed because there was no workflow!?!",
                    indent=4 * " ",
                    wrap=True,
                    dedent=False,
                )
            )
            return

        if not schema["workflow"][0]["success"]:
            text = "The Optimization step failed. There is no output at all!"
            printer.normal(__(text, indent=4 * " ", wrap=True, dedent=False))
            text = schema["workflow"][0]["error"]
            printer.normal(__(text, indent=4 * " ", wrap=False, dedent=False))
            raise RuntimeError(f"The TorchANI optimization failed:\n{text}")
            return

        P = self.parameters.current_values_to_dict(
            context=seamm.flowchart_variables._data
        )

        results = schema["systems"][0]["configurations"][0]["results"]["data"][step_no]

        results["energy,units"] = "eV"
        results["gradients,units"] = "eV/Å"

        energy = Q_(results["energy"], "eV").to("Eh")
        n_steps = results["number of optimization steps"]

        force_units = P["convergence"].units

        gradients = results["gradients"]
        max_derivative = 0
        rms = 0.0
        for row in gradients:
            sum = 0.0
            for v in row:
                sum += v**2
            dE = math.sqrt(sum)
            rms += sum
            if abs(dE) > max_derivative:
                max_derivative = abs(dE)
        rms = Q_(math.sqrt(rms / len(gradients)), "eV/Å").to(force_units)
        max_derivative = Q_(max_derivative, "eV/Å").to(force_units)

        text = ""
        if table is None:
            table = {
                "Property": [],
                "Value": [],
                "Units": [],
            }

        text = "The optimization converged."

        table["Property"].append("Steps")
        table["Value"].append(n_steps)
        table["Units"].append("")

        table["Property"].append("Total Energy")
        table["Value"].append(f"{energy.magnitude:.6f}")
        table["Units"].append("E_h")

        table["Property"].append("Maximum Force")
        table["Value"].append(f"{max_derivative.magnitude:.4f}")
        table["Units"].append(str(max_derivative.units))

        table["Property"].append("RMS Force")
        table["Value"].append(f"{rms.magnitude:.4f}")
        table["Units"].append(str(rms.units))

        tmp = tabulate(
            table,
            headers="keys",
            tablefmt="rounded_outline",
            colalign=("center", "decimal", "left"),
            disable_numparse=True,
        )
        length = len(tmp.splitlines()[0])
        text_lines = []
        text_lines.append("Results".center(length))
        text_lines.append(tmp)

        if text != "":
            text = str(__(text, indent=self.indent + 4 * " "))
            text += "\n\n"
        text += textwrap.indent("\n".join(text_lines), self.indent + 7 * " ")
        printer.normal(text)

        # Get the appropriate system/configuration for the new coordinates
        system, configuration = self.get_system_configuration(P)

        if P["system name"] == "optimized with <Hamiltonian>":
            system.name = f"optimized with {P['model']}"
        if P["configuration name"] == "optimized with <Hamiltonian>":
            configuration.name = f"optimized with {P['model']}"

        configuration.atoms.set_coordinates(results["coordinates"])

        text = (
            f"\nThe optimized structure has been placed in {system.name}/"
            f"{configuration.name}"
        )

        printer.normal(__(text, indent=self.indent + 4 * " "))

        # Citation for the specific method
        self.references.cite(
            raw=self._bibliography[P["model"]],
            alias="P['model']",
            module="torchani_step",
            level=2,
            note=f"The citation for the {P['model']} ML model.",
        )
        self.references.cite(
            raw=self._bibliography["ase"],
            alias="ase",
            module="torchani_step",
            level=2,
            note="The citation for ASE.",
        )
